import logging
import toml
import uvicorn
from threading import Thread
from camera_service.cam_capture import cam_capture
from twirp.asgi import TwirpASGIApp
from proto_services.camera_service_twirp import CameraServiceServer
import proto_services.camera_service_pb2 as camera_service
from camera_service.data.frames import modules_result
from camera_service.modules.modules_list import modules
from starlette.middleware.cors import CORSMiddleware

from camera_service.helpers.poetry_install import poetry_install
from camera_service.generate_code import generate_code
from google.protobuf.empty_pb2 import Empty
from camera_service.constants import PORT
from camera_service.utils.get_local_ip import get_local_ip

class CameraService(object):
    def IsCameraAlive(self, ctx, request):
        return camera_service.IsCameraAliveResponse(is_alive=True)
        

    def InstallModules(self, ctx, request):
        # Add packages from array to the project.
        with open('pyproject.toml', 'r') as file:
            pyproject_data = toml.load(file)

        for package in request.modules:
            pyproject_data['tool']['poetry']['dependencies'][package.name] = package.version

        with open('pyproject.toml', 'w') as file:
            toml.dump(pyproject_data, file)

        generate_code()

        return Empty()
    
    def UninstallModules(self, ctx, request):
        # Remove packages from array from the project.
        original_names = []

        #find original_name of the package
        for module in request.modules:
            for installed_module in modules:
                if installed_module.get("name") == module:
                    original_names.append(installed_module.get("package_name"))

        with open('pyproject.toml', 'r') as file:
            pyproject_data = toml.load(file)

        for package in original_names:
            del pyproject_data['tool']['poetry']['dependencies'][package]

        with open('pyproject.toml', 'w') as file:
            toml.dump(pyproject_data, file)

        generate_code()

        return Empty()
    
    def GetInstalledModules(self, ctx, request):
        formatted_modules = []

        for module in modules:
            formatted_modules.append(camera_service.InstalledModule(name = module.get("name"), options = module.get("options")))

        return camera_service.InstalledModules(modules=formatted_modules)

    def GetLastFrame(self, ctx, request):
        module_result = modules_result.get(request.module)
        if module_result is None or module_result.get("ok") == False:
            raise Exception(f"Module {request.module} failed processing")
        
        frame = module_result.get("frame")

        return camera_service.Frame(frame=frame)

app = TwirpASGIApp()
service = CameraServiceServer(service=CameraService())
app.add_service(service)

# Wrap the app with CORSMiddleware
app = CORSMiddleware(
    app,
    allow_origins=["*"],
    allow_methods=["*"],
    allow_headers=["*"],
)

def run_server():
    uvicorn.run(app=app, host=get_local_ip(), port=PORT)

def start():
    logging.basicConfig()

    # Install packages.
    poetry_install()

    thread = Thread(target = run_server, args = ())
    thread.start()
    cam_capture()



import cv2


def test_module_processing(frame, frame_number):
	center_x = frame.shape[1] // 2
	center_y = frame.shape[0] // 2

	# Draw a circle at the center
	radius = 50
	color = (0, 255, 0)  # BGR color (here, green)
	thickness = 2
	cv2.circle(frame, (center_x, center_y), radius, color, thickness)
	return {"frame": frame}


def original_frame_processing(frame, frame_number):
	return

modules = [
{ "name": "Original Frame", "package_name": "original", "proccessing": original_frame_processing },
{ "name": "test Module", "package_name": "test_module", "proccessing": test_module_processing },
{ "name": "test Module2", "package_name": "test_module2", "proccessing": original_frame_processing, "options": { "processing_frame": 10 } }
]


from camera_service.utils.get_modules_from_toml import get_modules_from_toml

def generate_code():
    modules = get_modules_from_toml()
    import_code = []
    original_frame_processing = "def original_frame_processing(frame, frame_number):\n\treturn"
    modules_list = ['{ "name": "Original Frame", "package_name": "original", "proccessing": original_frame_processing }']



    for package, version in modules:
        import_statement = f"import {package}"
        import_code.append(import_statement)
        module_item = f"""{{ "name": {package}.name, "package_name": "{package}", "proccessing": {package}.proccessing, "options": {package}.options }}"""

        modules_list.append(module_item)

    with open('.\\camera_service\\modules\\modules_list.py', 'w') as out_file:
        out_file.write('\n'.join([*import_code, "\n", original_frame_processing, "\n" "modules = [", ", ".join(modules_list), "]"]))

if __name__ == "__main__":
    generate_code()


import cv2
from camera_service.data.frames import modules_result
from camera_service.modules.modules_processing import modules_processing

def cam_capture():
    cap = cv2.VideoCapture(0)  # 0 for default webcam, change if necessary
    frame_number = 0
    while True:
        try:
            ret, frame = cap.read()
            if not ret:
                break
            frame_number += 1

            modules_processing(frame, frame_number)
            # cv2.imshow('Webcam', frame)

            # if cv2.waitKey(1) & 0xFF == ord('q'):
                # break

        except Exception as e:
            print(f"Error processing frame {frame_number}: {e}")
    cap.release()
    cv2.destroyAllWindows()

if __name__ == "__main__":
    cam_capture()


import toml
from camera_service.constants import PATTERN

def get_modules_from_toml():
    with open('pyproject.toml', 'r') as file:
        pyproject_data = toml.load(file)

    if 'tool' in pyproject_data and 'poetry' in pyproject_data['tool']:
        dependencies = pyproject_data['tool']['poetry'].get('dependencies', {})

    # filter out the modules that start with the PATTERN
    modules = [module for module in dependencies.items() if module[0].startswith(PATTERN)]

    return modules


from camera_service.modules.modules_list import modules
from camera_service.data.frames import modules_result
import concurrent.futures
import cv2

# Define a thread pool
pool = concurrent.futures.ThreadPoolExecutor()

# Dictionary to hold futures for each module
module_futures = {}


def modules_processing(originalFrame, frame_number):
	frameCopy = originalFrame.copy()

	# Submit each module processing task to the thread pool
	for module in modules:
		previous_module_future = module_futures.get(module.get("name"))
		skip_frames = None
		if module and module.get("options"):
			skip_frames = module.get("options").get("processing_frame")

		# Check if the previous module is done processing
		if previous_module_future is None or (skip_frames is None and previous_module_future.running() is False) or (skip_frames is not None and frame_number % skip_frames == 0):
			# Submit module processing task to the thread pool and store the future

			future = pool.submit(module.get("proccessing"), frameCopy, frame_number)	
			module_futures[module.get("name")] = future
			future.add_done_callback(lambda f, moduleName=module.get("name"): store_result(f, moduleName, originalFrame))

def store_result(future, module_name, original_frame):
	try:
		result = future.result()
		# Get the frame from the result or use the original frame
		frame = result.get("frame") if result is not None and result.get("frame") is not None else original_frame
		modules_result[module_name] = {
			"ok": True,
			# convert to jpeg and save in variable
			"frame": cv2.imencode('.jpg', frame)[1].tobytes()
		}
	except Exception as e:
		modules_result[module_name] = {"ok": False}
		print(f"Error store frame in {module_name} module: {e}")


import subprocess

def poetry_install():
    command = ['poetry', 'install']

    try:
        result = subprocess.run(command, capture_output=True, text=True, check=True)
        print(result.stdout)
    except subprocess.CalledProcessError as e:
        print(f"Error: {e.stderr}")